import math

noop = lambda i: i

def paginate(items, per_page=10, process=noop, extra={}, query_dict={}):
    page = int(query_dict.get('page') or 1)
    per_page = int(query_dict.get('per_page') or 0) or per_page
    offset = (page - 1) * per_page
    total = len(items)
    return {
        'pages': math.ceil(total / per_page),
        'items': [process(i) for i in items[offset: offset + per_page]],
        'page': page,
        'total': len(items),
        'next_page': page + 1 if offset + per_page < total else None,
        'prev_page': page - 1 if page > 1 else None,
        **extra,
    }


from collections import defaultdict
import imagehash
from io import BytesIO
import json
import numpy as np
import os
from PIL import Image
import urllib
import cv2


def fail(reason):
  def func(*args, **kwargs):
    raise Exception(reason)
  return func

try:
  from django.core.files.base import ContentFile
except ImportError:
  Contentfile = fail("You must install django.")

from .max_rect import max_rect

def _get_format(img):
  if isinstance(img, np.ndarray):
    return "np"
  return "pil"

def _coerce(img, format_):
  if type(img) == str:
    if img.startswith('data:'):
      if format_ == 'dataurl':
        # currently the only way to output a dataurl is this idempotent action
        return format_
      with urllib.request.urlopen(img) as response:
        with open('/tmp/sprite.png', 'wb') as f:
          f.write(response.read())
        img = Image.open('/tmp/sprite.png', mode="RGBA")

    else:
      img = Image.open(img).convert('RGBA')

  if format_ == "np":
    return np.array(img)

  if format_ == "pil":
    if isinstance(img, Image.Image):
      return img
    if isinstance(img, np.ndarray):
      mode = 'RGBA' if img.shape[-1] == 4 else 'RGB'
      return Image.fromarray(img, mode=mode)
    raise ValueError(f"Unknown image source of type: {type(img)}")

  raise ValueError(f"Unknown image format requested: {format_}")


def make_content_file(img, name):
  img = _coerce(img, "pil")
  img_io = BytesIO()
  img.save(img_io, format=name.split('.')[-1])
  return ContentFile(img_io.getvalue(), name)


def replace_color(data, color1, color2):
  r1, g1, b1 = color1
  r2, g2, b2 = color2

  red, green, blue = data[:,:,0], data[:,:,1], data[:,:,2]
  mask = (red == r1) & (green == g1) & (blue == b1)
  data[:,:,:3][mask] = [r2, g2, b2]

  return data


def paste_image(back, front, x, y):
  # replaces the pixels of front with back, treating alpha like just another channel
  bh, bw = back.shape[:2]
  fh, fw = front.shape[:2]
  x1, x2 = max(x, 0), min(x+fw, bw)
  y1, y2 = max(y, 0), min(y+fh, bh)
  front_cropped = front[y1-y:y2-y, x1-x:x2-x]
  back_cropped = back[y1:y2, x1:x2]

  result = back.copy()
  result[y1:y2, x1:x2] = front_cropped
  return result


def merge_image(back, front, x, y):
  # like paste image but respects alpha channel (if it exists)
  needs_revert = False
  if back.shape[2] == 3:
    needs_revert = True
    back = cv2.cvtColor(back, cv2.COLOR_BGR2BGRA)
  if front.shape[2] == 3:
    front = cv2.cvtColor(front, cv2.COLOR_BGR2BGRA)

  # crop the overlay from both images
  bh,bw = back.shape[:2]
  fh,fw = front.shape[:2]
  x1, x2 = max(x, 0), min(x+fw, bw)
  y1, y2 = max(y, 0), min(y+fh, bh)
  front_cropped = front[y1-y:y2-y, x1-x:x2-x]
  back_cropped = back[y1:y2, x1:x2]

  alpha_front = front_cropped[:,:,3:4] / 255
  alpha_back = back_cropped[:,:,3:4] / 255

  # replace an area in result with overlay
  result = back.copy()
  result[y1:y2, x1:x2, :3] = alpha_front * front_cropped[:,:,:3] + (1-alpha_front) * back_cropped[:,:,:3]
  result[y1:y2, x1:x2, 3:4] = (alpha_front + alpha_back) / (1 + alpha_front*alpha_back) * 255

  if needs_revert:
    result = cv2.cvtColor(result, cv2.COLOR_BGRA2BGR)
  return result


def get_or_create(image_name, function, path="", force=False):
  target_path = os.path.join(path, image_name)
  if not force and os.path.exists(target_path):
    return Image.open(target_path), False

  image = function()
  image.save(target_path)
  return image, True


# ideally we should do something more like this: https://stackoverflow.com/a/43111221/266564
def analyze_colors(img, ignore_clear=True):
  img = _coerce(img, 'np')
  counts = defaultdict(int)
  ignore_clear = ignore_clear and img.shape[2] == 4
  all_colors = []
  for row in img:
    for color in row:
      if ignore_clear and color[3] == 0:
        continue
      counts[tuple(color)] += 1
      all_colors.append(color[:3])
  average = np.round(np.sum(all_colors, axis=0) / len(all_colors))
  return { 'counts': counts, 'average': average }

def int_to_64bit_array(value):
  array = [int(i) for i in bin(value)[2:]]
  while len(array) < 64:
    array.insert(0, 0)
  return array


def int_to_imagehash(value):
  if isinstance(value, imagehash.ImageHash):
    return value
  hash_ = np.array(int_to_64bit_array(value)).reshape((8, 8))
  return imagehash.ImageHash(hash_)


def imagehash_to_int(value):
  if isinstance(value, int):
    return value
  res = 0
  for row in value.hash:
    for boolean in row:
      res = (res << 1) | int(boolean)
  return res


def color_distance(color1, color2):
  return np.sum(np.abs(np.array(color1) - np.array(color2)))


class NpEncoder(json.JSONEncoder):
  def default(self, obj):
    if isinstance(obj, np.integer):
      return int(obj)
    if isinstance(obj, np.floating):
      return float(obj)
    if isinstance(obj, np.ndarray):
      return obj.tolist()
    return super().default(obj)


def make_holes(image, holes, color=(0,0,0,0)):
    format_ = _get_format(image)
    image = _coerce(image, 'np')
    for x, y in holes:
        image[y*256:(y+1) * 256,x*256:(x+1) * 256,:] = [0,0,0,0]
    return _coerce(image, format_)



from unrest.utils import mkdir

import cv2
import sys
import os

def white_background(image):
  image = image.copy()
  trans_mask = image[:,:,3] == 0
  image[trans_mask] = [255, 255, 255, 255]
  return cv2.cvtColor(image, cv2.COLOR_BGRA2BGR)

def extract_sprites(image, dest='.cache', show=True):
  fname = image.rsplit('.', 1)[0]
  dest = mkdir(dest, fname)
  if isinstance(image, str):
    image = cv2.imread(image, cv2.IMREAD_UNCHANGED)
  white = white_background(image)
  gray = cv2.cvtColor(white, cv2.COLOR_BGR2GRAY)
  thresh = cv2.threshold(gray, 0, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)[1]

  # Original script used the following to try to get multipart sprites to join together
  # maybe make this a kwarg option?
  # kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (3,3))
  # close = cv2.morphologyEx(thresh, cv2.MORPH_CLOSE, kernel, iterations=2)
  # dilate = cv2.dilate(close, kernel, iterations=1)

  cnts = cv2.findContours(thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
  cnts = cnts[0] if len(cnts) == 2 else cnts[1]

  sprite_number = 0
  crops = []
  for c in cnts:
    x, y, w, h = cv2.boundingRect(c)
    ROI = image[y:y+h, x:x+w]
    out_path = os.path.join(dest, f'{fname}_{sprite_number}.png')
    crops.append(out_path)
    cv2.imwrite(out_path, ROI)
    cv2.rectangle(image, (x, y), (x + w, y + h), (36, 255, 12), 2)
    sprite_number += 1

  if show:
    cv2.imshow('thresh', thresh)
    # cv2.imshow('dilate', dilate)
    cv2.imshow('image', image)
    cv2.waitKey()


if __name__ == "__main__":
  extract_sprites(sys.argv[1])


# adapted from https://stackoverflow.com/a/30418912/266564

import numpy

def max_rect(array):
    nrows, ncols = array.shape
    skip = 0
    area_max = (0, [])

    w = numpy.zeros(dtype=int, shape=array.shape)
    h = numpy.zeros(dtype=int, shape=array.shape)
    for r in range(nrows):
        for c in range(ncols):
            if array[r][c] == skip:
                continue
            if r == 0:
                h[r][c] = 1
            else:
                h[r][c] = h[r-1][c]+1
            if c == 0:
                w[r][c] = 1
            else:
                w[r][c] = w[r][c-1]+1
            minw = w[r][c]
            for dh in range(h[r][c]):
                minw = min(minw, w[r-dh][c])
                area = (dh+1)*minw
                if area > area_max[0]:
                    area_max = (area, [(r-dh, c-minw+1, r, c)])

    if not area_max[1]:
        return [0, 0, 0, 0,]

    y1, x1, y2, x2 = area_max[1][0]
    return [min(x1, x2), min(y1, y2), abs(x2-x1)+1, abs(y2-y1)+1]

if __name__ == "__main__":
    s = '''
1 1 1 1 0 1
1 1 0 1 1 0
1 1 1 1 1 1
0 1 1 1 1 1
1 1 1 1 1 0
1 1 1 0 1 1'''
    nrows = 6
    ncols = 6
    array = numpy.fromstring(s, dtype=int, sep=' ').reshape(nrows, ncols)
    x, y, w, h = max_rect(array)
    print(x, y, w, h)
    assert(x == 1)
    assert(y == 2)
    assert(w == 4)
    assert(h == 3)


from django.http import JsonResponse, HttpResponseRedirect, HttpResponse
from django.contrib.auth import login, get_user_model
from django.contrib.auth.forms import PasswordResetForm
from django.contrib.auth.tokens import default_token_generator
from django.core.exceptions import ValidationError
from django.urls import reverse
from django.utils.http import urlsafe_base64_decode

import json
import random


def bad_token(request):
    message = "The link you are using is invalid or expired. Please try again."
    return HttpResponse("<h1>Bad Token</h1><p>{}</p>".format(message))


def create(request):
    data = json.loads(request.body.decode("utf-8") or "{}")
    username = email = data.get("email",None)
    User = get_user_model()
    if email and User.objects.filter(email=email):
        raise NotImplemented()
    while not username or User.objects.filter(username=username):
        _hash = "{:032x}".format(random.getrandbits(128))
        username = "guest-{}".format(_hash[:8])
        email = username + "@example.com"
    user = User.objects.create(
        username=username,
        email=email
    )
    user.backend = "django.contrib.auth.backends.ModelBackend"
    login(request,user)
    return JsonResponse({"status": "ok"})


def change_email(request):
    data = json.loads(request.body.decode("utf-8"))
    email = data.get("email",None)
    User = get_user_model()
    user = request.user
    if User.objects.exclude(id=user.id).filter(email=email):
        return JsonResponse(
            {'error': 'Another account already has that email.'},
            status=400
        )
    user = request.user
    user.email = user.username = email
    user.save()
    return JsonResponse({"success": "Email address updated"})


def send(request):
    data = json.loads(request.body.decode("utf-8"))
    form = PasswordResetForm(data)
    if form.is_valid():
        form.save(
            subject_template_name='email/nopass/subject.txt',
            email_template_name='email/nopass/body.txt',
        )
        return JsonResponse({"success":"Please check your email for a login link."})
    return JsonResponse({},status=400)


# django.contrib.auth.views.PasswordResetConfirmView.get_user
def get_user(uidb64):
    User = get_user_model()
    try:
        # urlsafe_base64_decode() decodes to bytestring
        uid = urlsafe_base64_decode(uidb64).decode()
        return User.objects.get(pk=uid)
    except (TypeError, ValueError, OverflowError, User.DoesNotExist, ValidationError):
        return None


# adapted from django.contrib.auth.views.PasswordResetConfirmView
def complete_login(request, uidb64=None, token=None):
    redirect_url = '/'
    if request.user.is_authenticated:
        return HttpResponseRedirect(redirect_url)
    user = get_user(uidb64)

    if user and default_token_generator.check_token(user, token):
        # log the user and redirect to change password
        # logging in invalidates the token
        user.backend = "django.contrib.auth.backends.ModelBackend"
        login(request, user)

        # this forces them to change password
        user.set_unusable_password()
    else:
        redirect_url = reverse("nopass_bad_token")
    return HttpResponseRedirect(redirect_url)


import json
import random
from django.http import JsonResponse
from django.conf import settings
from django.contrib.auth import login, logout, get_user_model

# This registers all the form views
from unrest.user.forms import get_reset_user

def user_json(request):
    user = request.user
    if not user.is_authenticated:
        return JsonResponse({})
    keys = ['id', 'username', 'email', 'is_superuser', 'is_staff']
    data = { k: getattr(user,k) for k in keys }
    if hasattr(user_json, 'get_extra'):
        data.update(user_json.get_extra(user))
    for func in user_json.extras:
        data.update(func(request))
    return JsonResponse({ 'user': data })

def signup_guest(request):
    if not getattr(settings, 'UNREST_ALLOW_GUEST', False):
        return JsonResponse(status=400)
    User = get_user_model()
    rando = random.randint(0, 100000)
    while User.objects.filter(username=f'guest{rando}'):
        rando = random.randint(0, 100000)
    user = User.objects.create(username=f'guest{rando}')
    login(request, user, backend='django.contrib.auth.backends.ModelBackend')
    return JsonResponse({})

user_json.extras = []

def logout_ajax(request):
  logout(request)
  return JsonResponse({})

def password_reset(request, uidb64, token):
    user = get_reset_user(uidb64, token)
    if not user:
        return HttpResponseRedirect("/#/reset-unsuccessful/") # TODO
    request.session['reset-uidb64'] = uidb64
    request.session['reset-token'] = token
    return HttpResponseRedirect('/#/reset/')


import cv2
import numpy as np
import sys
import unrest.image as img

def list_colors(image, ignore_black=True, cutoff=0.01):
    if isinstance(image, str):
        image = cv2.imread(image)
    colors = image.reshape(-1, image.shape[-1])
    counts = np.unique(colors, axis=0, return_counts = True)
    counts = sorted(zip(*counts),key=lambda a: -a[1])
    if ignore_black:
        counts = [c for c in counts if any(c[0])]
    if cutoff:
        total = sum([c[1] for c in counts])
        counts = [c for c in counts if c[1] > total * cutoff]
    return counts

if __name__ == "__main__":
    colors = list_colors(sys.argv[1])
    image = cv2.imread(sys.argv[1])
    cv2.imshow('og', image)
    result = None
    for color, count in colors[:4]:
        reduced = cv2.inRange(image, color, color)
        cv2.imshow(str(color), reduced)
        if result is not None:
            result = result | reduced
        else:
            result = reduced
    cv2.imshow('final', result)
    cv2.waitKey(0)


import random

from django.conf import settings
from django import forms
from django.contrib.auth import login, get_user_model
from django.contrib.auth.forms import PasswordResetForm, SetPasswordForm
from django.contrib.auth.tokens import default_token_generator
from django.utils.http import urlsafe_base64_decode

if 'django_registration' in settings.INSTALLED_APPS:
    from django_registration.forms import RegistrationFormUniqueEmail
    from django_registration.backends.activation.views import RegistrationView
    from django_registration import validators
else:
    class RegistrationFormUniqueEmail(forms.ModelForm):
        class Meta:
            pass

import unrest_schema

def get_reset_user(uidb64, token):
    User = get_user_model()
    try:
        uid = urlsafe_base64_decode(uidb64).decode()
        user = User.objects.get(pk=uid)
    except (TypeError, ValueError, OverflowError, User.DoesNotExist, ValidationError):
        return None

    if default_token_generator.check_token(user, token):
        return user

@unrest_schema.register
class PasswordResetForm(PasswordResetForm):
    user_can_POST = 'ANY'
    # the django password reset form uses a bunch of kwargs on save, making it very non-standard
    # we hack them in here so that this plays nice with the rest of the schema form flow
    def save(self, *args, **kwargs):
        kwargs['request'] = self.request
        return super().save(*args, **kwargs)

@unrest_schema.register
class SetPasswordForm(SetPasswordForm):
    user_can_POST = 'ANY'
    # In django, token validation is done in the view and user is passed into the form
    # this does all that in clean instead to make it fit into schema form flow
    def __init__(self, *args, **kwargs):
        super().__init__(None, *args, **kwargs)
        del self.fields['new_password1'].help_text

    def clean(self):
        uidb64 = self.request.session.get('reset-uidb64', '')
        token = self.request.session.get('reset-token', '')
        self.user = get_reset_user(uidb64, token)
        if not self.user:
            raise forms.ValidationError('This password reset token has expired')
        return self.cleaned_data

    def save(self, commit=True):
        # password reset token is invalid after save. Remove from session
        user = super().save(commit)
        self.request.session.pop('reset-uidb64', None)
        self.request.session.pop('reset-token', None)
        login(self.request, user, backend='django.contrib.auth.backends.ModelBackend')
        return user

@unrest_schema.register
class SignUpForm(RegistrationFormUniqueEmail):
    password1 = forms.CharField(label='Password', max_length=128, widget=forms.PasswordInput)
    user_can_POST = 'ANY'
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['username'].help_text = None
        self.fields['password1'].help_text = None
        self.fields.pop('password2')
    def clean(self, *args, **kwargs):
        self.cleaned_data['password2'] = self.cleaned_data.get('password1')
        super().clean()
    class Meta(RegistrationFormUniqueEmail.Meta):
        model = get_user_model()
        fields = ['username', 'email', 'password1']

    def save(self, commit=False):
        user = super().save(commit=False)

        if getattr(settings, 'UNREST_VERIFY_EMAIL', None):
            # use django_registration to send email to user
            user.is_active = False

            view = RegistrationView()
            view.request = self.request
            view.send_activation_email(user)
            user.save()
        else:
            user.save()
            login(self.request,  user, backend='django.contrib.auth.backends.ModelBackend')

        return user


@unrest_schema.register
class LoginForm(forms.Form):
    user_can_POST = 'ANY'
    username = forms.CharField(label='Username', max_length=150)
    password = forms.CharField(label='Password', max_length=128, widget=forms.PasswordInput)
    def clean(self):
        User = get_user_model()
        username = self.cleaned_data.get('username')
        password = self.cleaned_data.get('password')
        if not username and password:
            return self.cleaned_data
        user = User.objects.filter(username=username).first()
        if user:
            if user.check_password(password):
                self.user = user
                return self.cleaned_data
        raise forms.ValidationError("Username and password do not match", code='password_mismatch')
    def save(self, commit=True):
        login(self.request, self.user, backend='django.contrib.auth.backends.ModelBackend')


@unrest_schema.register
class UserSettingsForm(forms.ModelForm):
    user_can_GET = 'SELF'
    user_can_PUT = 'SELF'
    def clean_username(self):
        value = self.cleaned_data.get('username')
        if value != self.request.user.username:
            error_message = 'Another user has this username.'
            validators.validate_unique(get_user_model(), 'username', error_message)(value)
        validators.ReservedNameValidator()(value)
        return value
    def clean_email(self):
        value = self.cleaned_data.get('email')
        if value != self.request.user.email:
            error_message = 'Another user has this email.'
            validators.validate_unique(get_user_model(), 'email', error_message)(value)
        return value
    class Meta:
        model = get_user_model()
        fields = getattr(settings, 'UNREST_USER_SETTINGS_FIELDS', ['username', 'email'])


# usage: (in project settings file)
# from unrest.settings import get_secret_key
# SECRET_KEY = get_secret_key(BASE_DIR)

import os
from django.core.management.utils import get_random_secret_key

def get_secret_key(BASE_DIR):
    key_path = os.path.join(BASE_DIR, '.secret_key')
    if os.path.exists(key_path):
        with open(key_path, 'r') as f:
            SECRET_KEY = f.read()
    else:
        SECRET_KEY = get_random_secret_key()
        with open(key_path, 'w') as f:
            f.write(SECRET_KEY)
            print('wrote secret key to', key_path)

    return SECRET_KEY


from django.apps import apps
from django.conf import settings
from django.contrib.staticfiles import finders
from django.http import JsonResponse, HttpResponseRedirect, HttpResponse
from django.views.decorators.csrf import ensure_csrf_cookie
from django.views.static import serve
from django.middleware.csrf import get_token
import json
import subprocess
import os

@ensure_csrf_cookie
def index(request, *args, **kwargs):
    path = kwargs.get('path', finders.find('index.html'))
    response = serve(
        request,
        os.path.basename(path),
        os.path.dirname(path)
    )
    _hash = subprocess.check_output(['git', 'rev-parse', 'HEAD'])
    response.set_cookie("GIT_HASH", _hash.decode('utf-8').strip())
    return response

@ensure_csrf_cookie
def get_csrf(request):
    return HttpResponse(get_token(request))


def spa(request, *args, **kwargs):
    raise Exception('Use index instead of spa')


def favicon(request):
    path = finders.find(getattr(settings,'FAVICON','favicon.ico'))
    if not path:
        return HttpResponse(status=404)
    return serve(
        request,
        os.path.basename(path),
        os.path.dirname(path)
    )

def redirect(request,url=None):
    return HttpResponseRedirect(url)

def superuser_api_view(request, app_name, model_name):
    app = apps.get_app_config(app_name)
    model = app.get_model(model_name)
    data = json.loads(request.body.decode('utf-8') or "{}")
    if not request.method == "POST":
        return list_view(request, app_name, model_name)
    if not request.user.is_superuser:
        raise NotImplementedError("Only superusers can use this view")
    data = json.loads(request.body.decode('utf-8') or "{}")
    id = data.pop("id", 0)
    if id:
        obj = model.objects.get(id=id)
        obj.data = data
        obj.save()
    else:
        obj = model.objects.create(data=data)
    return JsonResponse(obj.as_json)


def list_view(request, app_name, model_name):
    app = apps.get_app_config(app_name)
    model = app.get_model(model_name)
    data = json.loads(request.body.decode('utf-8') or "{}")
    if data:
        id = data.pop("id",None)
        if id:
            obj = model.from_data(data,request=request,id=id)
        else:
            obj = model.from_data(data,request=request)
        obj.save()
        return JsonResponse(obj.as_json)
    items = model.objects.request_filter(request)
    return JsonResponse({
        'results': [i.as_json for i in items],
    })

def intentional_500(request, *args, **kwargs):
    raise Exception("Intentional 500 error")


import cv2
import numpy as np
from pathlib import Path
from unrest.image import merge_image, paste_image

DIR = Path(__file__).parent.absolute()

def test_merge_image():
    back = cv2.imread(str(DIR / 'lz.png'))
    ball = cv2.imread(str(DIR / 'morphball.png'), cv2.IMREAD_UNCHANGED)
    print(back.shape)
    print(ball.shape)
    result = merge_image(back, ball, 50, 50)
    cv2.imwrite('result_merge.png', result)


def test_add():
    result = cv2.imread(str(DIR / 'lz.png'))
    ball = cv2.imread(str(DIR / 'morphball.png'))
    for i in range(10):
        empty = result.copy()
        empty[:,:,:] = 0
        empty = merge_image(empty, ball, i * 16, 70)
        result = result + (empty * (i+1))
    cv2.imwrite('result_add.png', result)


def test_overflow_add():
    result = cv2.imread(str(DIR / 'lz.png'))
    result = result.astype('int32')
    ball = cv2.imread(str(DIR / 'morphball.png')).astype('int32')
    for i in range(20):
        empty = result.copy()
        empty[:,:,:] = 0
        empty = paste_image(empty, ball, i * 16, 70)
        result = result + (empty * (i+1))
    mask = result[:,:,:] > 255
    print(np.amax(result))
    result[mask] = 255
    cv2.imwrite('result_overflow.png', result)
    arst



from django.conf import settings
from django.urls import path, re_path, include

from unrest.views import index, favicon, intentional_500

urlpatterns = [
    path('intentional_500/', intentional_500),
    path('favicon.ico', favicon),
    re_path('^(?:login|logout|signup|reset-password|new)/', index),
    re_path('^$', index),
    re_path('', include('unrest.user.urls')),
    re_path('', include('unrest_schema.urls')),
]

if settings.DEBUG:  # pragma: no cover
    from django.views.static import serve
    urlpatterns += [
        re_path(r'^media/(?P<path>.*)$', serve, {
            'document_root': settings.MEDIA_ROOT,
            'show_indexes': True
        }),
    ]


from functools import wraps
import os
import json
import time

def mkdir(root, *args):
    parts = os.path.join(*args).strip('/').split('/')
    path = root
    if not os.path.exists(path):
        os.mkdir(path)
    for part in parts:
        path = os.path.join(path, part)
        if not os.path.exists(path):
            os.mkdir(path)
    return path


class JsonCache(dict):
    def __init__(self, path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._path = path
        if os.path.exists(self._path):
            with open(self._path, 'r') as f:
                self.update(json.loads(f.read()))
    def __setitem__(self, key, value):
        if callable(value):
            if not key in self:
                super().__setitem__(key, value())
        else:
            super().__setitem__(key, value)
        self._save()
    def _save(self):
        with open(self._path, 'w') as f:
            f.write(json.dumps(self, indent=2))

def _ms(seconds):
    if seconds <10:
        return round(seconds * 1000,3)
    return int(seconds * 1000)

def time_it(func, verbose=False):
    if func == True:
        return lambda func: time_it(func, verbose=True)
    times = []
    @wraps(func)
    def wrapped(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        dt = time.time()-start
        times.append(dt)
        if verbose:
            ave = _ms(sum(times) / len(times))
            _sum = _ms(sum(times))
            print(func.__name__, _ms(dt), ave, _sum)
        return result
    def _print():
        ave = _ms(sum(times) / len(times))
        _sum = _ms(sum(times))
        print(f'{func.__name__} ave={ave} total={_sum} count={len(times)}')
    wrapped.print = _print
    return wrapped

class Ticker:
    def __init__(self, name=None):
        self.name = name
        self.last = self.start = time.time()
        self.counts = 0
    def __call__(self):
        self.counts += 1
        _now = time.time()
        dt = _ms(_now - self.last)
        s = f'{name}\tdt={dt}'
        # if new_name == self.name:
        #     ave = _ms(_now - self.start) // self.counts)
        #     s += f'\tave={ave}\tcounts={counts}'
        # else:
        #     self.name = new_name
        #     self.counts = 0
        print(s)



from django.urls import path

from . import views

urlpatterns = [
    path("create/",views.create,name="nopass_create"),
    path("change_email/",views.change_email,name="nopass_change_email"),
    path("send/",views.send,name="nopass_send"),
    path("bad_token/",views.bad_token,name="nopass_bad_token"),
    path("<uidb64>/<token>/",views.complete_login,name="nopass_login"),
]


import functools

from django.http import JsonResponse


def login_required(func):
  @functools.wraps(func)
  def wrapped(request, *args, **kwargs):
    if request.user.is_authenticated:
      return func(request, *args, **kwargs)
    return JsonResponse({}, status_code=403)
  return wrapped

def cached_method(target):
  def wrapper(*args, **kwargs):
    obj = args[0]
    name = '___' + target.__name__

    if not hasattr(obj, name):
      value = target(*args, **kwargs)
      setattr(obj, name, value)

    return getattr(obj, name)

  return wrapper

def cached_property(target):
  return property(cached_method(target))



from django.contrib.auth import get_user_model
from django.test import TestCase
from django.urls import reverse
from django.core import mail

import json
import re

# Create your tests here.
class NopassTestCase(TestCase):
    def post_json(self,url,_dict):
        self.client.post(url,json.dumps(_dict),content_type="application/json")
    def test_user_flow(self):
        User = get_user_model()

        # create an anonymous user
        self.client.get(reverse("nopass_create"))
        self.assertEqual(User.objects.count(),1)
        user = User.objects.all()[0]
        self.assertEqual(user.username+"@example.com",user.email)

        # set user's email to something else
        data = {"email": "arst@asrt.com"}
        self.post_json(reverse("nopass_change_email"),data)
        self.assertEqual(User.objects.all()[0].email,data['email'])

        # get login link
        self.post_json(reverse("nopass_send"),data)
        self.assertEqual(len(mail.outbox),1)

        # login with url
        self.client.logout()
        self.assertTrue('_auth_user_id' not in self.client.session)
        body = mail.outbox[0].body
        url = body.split("http://example.com")[-1].split('\n')[0]
        response = self.client.get(url)
        self.assertEqual(self.client.session['_auth_user_id'],'1')

    def test_bad_token(self):
        redirect = self.client.get(reverse("nopass_login",args=['a','b']))
        self.assertTrue(redirect.url.endswith("bad_token/"))


#! TODO this was copied from lablackey.db.models, but much was deleted
# Named model, slug model, ordered model, tree model...
# review and decide which to keep

from django.conf import settings
from django.db import models
from django.db.models.fields.files import ImageFieldFile

from unrest.managers import RequestManager, UserRequestManager
import json

def as_json(self, json_fields=None):
  out = { f: getattr(self,f) for f in self.json_fields }
  out.update(**out.pop("data",{}))
  for f in self.fk_json_fields:
    if getattr(self,f):
      out[f] = getattr(self,f).as_json
  for f in self.m2m_json_fields:
    out[f] = [i.as_json for i in getattr(self,f)]
  return out

class JsonMixin(object):
  json_fields = []
  filter_fields = []
  m2m_json_fields = []
  fk_json_fields = []
  as_json = property(as_json)

_hash = lambda data: hash(json.dumps(data,sort_keys=True))

class JsonModel(models.Model, JsonMixin):
  created = models.DateTimeField(auto_now_add=True)
  data_hash = models.BigIntegerField()
  data = models.JSONField(default=dict)
  objects = RequestManager()

  def save(self,*args,**kwargs):
    self.data_hash = _hash(self.data)
    super().save(*args,**kwargs)

  @classmethod
  def from_data(cls,data,request=None,**kwargs):
    data_hash = _hash(data)
    instance = cls.objects.filter(data=data).first()
    return instance or cls.objects.create(
      data=data,
      data_hash=data_hash,
      **kwargs
    )

  class Meta:
    abstract = True

class UserModel(JsonModel):
  user = models.ForeignKey(settings.AUTH_USER_MODEL,on_delete=models.CASCADE)
  objects = UserRequestManager()
  class Meta:
    abstract = True

  @classmethod
  def from_data(cls,data,request=None,**kwargs):
    kwargs['user'] = request.user
    return super().from_data(data,request=None,**kwargs)


class BaseModel(models.Model):
  created = models.DateTimeField(auto_now_add=True)
  updated = models.DateTimeField(auto_now=True)
  class Meta:
    abstract = True

  def to_json(self, attrs=None):
    if attrs == None:
      attrs = self.json_fields
    result = {}

    for attr in attrs:
      value = getattr(self, attr)

      # fields can be callable, eg my_object.get_absolute_url()
      if callable(value):
        value = value()

      # Handle fields that are not serializeable by DjangoJSONEncoder
      if isinstance(value, ImageFieldFile):
        value = value.url
      result[attr] = value
    return result

def _choices(choices):
  processed = []
  for choice in choices:
    if not isinstance(choice, (list, tuple)):
      choice = [choice, choice]
    processed.append(choice)
  return processed


from django.db import models

class RequestManager(models.Manager):
    def request_filter(self,request,*args,**kwargs):
        return self.filter(*args,**kwargs)

class UserRequestManager(models.Manager):
    def request_filter(self,request,*args,**kwargs):
        if not request.user.is_authenticated:
            return self.none()
        kwargs['user'] = request.user
        return self.filter(*args,**kwargs)


from django.urls import path

from unrest.user import views
from unrest.views import spa

urlpatterns = [
    path('api/auth/user.json', views.user_json),
    path('api/auth/logout/', views.logout_ajax),
    path('api/auth/guest/', views.signup_guest),
    path('reset-password-confirm/<uidb64>/<token>/', spa, name='password_reset_confirm'),
]


from setuptools import setup, find_packages

setup(
  name = 'django-unrest',
  packages = find_packages(),
  version = '0.1.7',
  description = 'A collection of tools for django',
  long_description="",
  long_description_content_type="text/markdown",
  author = 'Chris Cauley',
  author_email = 'chris@lablackey.com',
  url = 'https://github.com/chriscauley/django-unrest',
  keywords = ['utils','django'],
  license = 'MIT',
  include_package_data = True,
  install_requires = [
    'social-auth-app-django>=4.0.0',
    'social-auth-core>=3.3.3'
  ]
)
